from polarization import *

# Define the vector with Ex and Ey
print("\n\nDefine horizontal polarization with JonesVector(Ex=1, Ey=0)")
print("===========================================================")
v = JonesVector(Ex=1, Ey=0) # horizontal
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()
#v.show("horizontal.mp4") #to save movie

print("\n\nDefine other polarizations with JonesVector(Ex, Ey),\ncan be normalized with .normalize()")
print("====================================================")
v = JonesVector(Ex=1, Ey=1).normalize() # +45°
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()
#v.show("plus45.mp4") #to save movie

print("\n\nMany predefined vectors exist")
print("=============================")
# Many predefined vectors:
# JonesVector.vertical()
# JonesVector.horizontal()
# JonesVector.plus45()
# JonesVector.minus45()
# JonesVector.rightCircular()
# JonesVector.leftCircular()

print("\nJonesVector.rightCircular()")
print("-----------------------------")
v = JonesVector.rightCircular() 
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()

print("\nJonesVector.leftCircular()")
print("-----------------------------")
v = JonesVector.leftCircular() # Many predefined vectors
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()

# Arbitrary polarization
print("\n\nArbitrary polarization JonesVector(Ex=1*exp(1j*0.3), Ey=0.5).normalize()")
print("========================================================================")
v = JonesVector(Ex=1*exp(1j*0.3), Ey=0.5).normalize()
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()

# Many properties are defined:
print("\n\nMany properties to access polarization state")
print("============================================")
print("For vector {0}".format(v))
print("v.orientation (0=horizontal) = {0}".format(v.orientation))
print("v.isLinearlyPolarized        = {0}".format(v.isLinearlyPolarized))
print("v.isEllipticallyPolarized    = {0}".format(v.isEllipticallyPolarized))
print("v.isCircularlyPolarized      = {0}".format(v.isCircularlyPolarized))
print("v.isRightCircularlyPolarized = {0}".format(v.isRightCircularlyPolarized))
print("v.isLeftCircularlyPolarized  = {0}".format(v.isLeftCircularlyPolarized))

# Vectors can be transformed by JonesMatrices
# Any matrix can be created with JonesMatrix(A,B,C,D)
# but there are many predefined matrices:
#
# HorizontalPolarizer(): polarizer at theta=0°
# VerticalPolarizer(): polarizer at theta=90° 
# Plus45Polarizer(): polarizer at theta=45°
# Minus45Polarizer(): polarizer at theta=-45°
# HWP(theta=pi/4) : halfwave plate at 45°
# QWP(theta=pi/4) : quarterwave plate at 45° 
# RightCircularPolarizer(): right circular polarizer
# LeftCircularPolarizer(): left circular polarizer
# PhaseRetarder(): arbitrary retarder

print("\n\nTransform the JonesVector with JonesMatrices")
print("============================================")
print("horizontal vector going through quarter waveplate")

vIn = JonesVector.horizontal()
v = QWP(theta=pi/4)*vIn

print("Input components are {0}".format(vIn))
print("Output components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
print("isCircular {0}".format(v.isCircularlyPolarized))
v.show()

print("\n\nApply several JonesMatrices sequentially")
print("============================================")

vIn = JonesVector.horizontal()
v = HWP(theta=pi/2)*QWP(theta=pi/3)*vIn

print("Input components are {0}".format(vIn))
print("Output components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
print("v.isLinearlyPolarized        = {0}".format(v.isLinearlyPolarized))
print("v.isEllipticallyPolarized    = {0}".format(v.isEllipticallyPolarized))
print("v.isCircularlyPolarized      = {0}".format(v.isCircularlyPolarized))
print("v.isRightCircularlyPolarized = {0}".format(v.isRightCircularlyPolarized))
print("v.isLeftCircularlyPolarized  = {0}".format(v.isLeftCircularlyPolarized))
v.show()

print("Malus law example")
print("=================")
x = []
y = []
vIn = JonesVector.horizontal()
for theta in range(0,190,10):
    pol = LinearPolarizer(theta=theta*radPerDeg)
    vOut = pol*vIn

    x.append(theta)
    y.append(vOut.intensity)

plt.title("\n\nMalus Law: horizontal beam intensity after linear polarizer")
plt.xlabel(r"Polarizer orientation $\theta$ from horizontal [°]")
plt.ylabel(r"Intensity [arb. unit]")
plt.plot(x,y,'ko')
plt.show()


print(r"""
# Example code for polarization module
# ====================================
from polarization import *

# Define the vector with Ex and Ey
print("\n\nDefine horizontal polarization with JonesVector(Ex=1, Ey=0)")
print("===========================================================")
v = JonesVector(Ex=1, Ey=0) # horizontal
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()
#v.show("horizontal.mp4") #to save movie

print("\n\nDefine other polarizations with JonesVector(Ex, Ey),\ncan be normalized with .normalize()")
print("====================================================")
v = JonesVector(Ex=1, Ey=1).normalize() # +45°
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()
#v.show("plus45.mp4") #to save movie

print("\n\nMany predefined vectors exist")
print("=============================")
# Many predefined vectors:
# JonesVector.vertical()
# JonesVector.horizontal()
# JonesVector.plus45()
# JonesVector.minus45()
# JonesVector.rightCircular()
# JonesVector.leftCircular()

print("\nJonesVector.rightCircular()")
print("-----------------------------")
v = JonesVector.rightCircular() 
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()

print("\nJonesVector.leftCircular()")
print("-----------------------------")
v = JonesVector.leftCircular() # Many predefined vectors
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()

# Arbitrary polarization
print("\n\nArbitrary polarization JonesVector(Ex=1*exp(1j*0.3), Ey=0.5).normalize()")
print("========================================================================")
v = JonesVector(Ex=1*exp(1j*0.3), Ey=0.5).normalize()
print("Components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
v.show()

# Many properties are defined:
print("\n\nMany properties to access polarization state")
print("============================================")
print("For vector {0}".format(v))
print("v.orientation (0=horizontal) = {0}".format(v.orientation))
print("v.isLinearlyPolarized        = {0}".format(v.isLinearlyPolarized))
print("v.isEllipticallyPolarized    = {0}".format(v.isEllipticallyPolarized))
print("v.isCircularlyPolarized      = {0}".format(v.isCircularlyPolarized))
print("v.isRightCircularlyPolarized = {0}".format(v.isRightCircularlyPolarized))
print("v.isLeftCircularlyPolarized  = {0}".format(v.isLeftCircularlyPolarized))

# Vectors can be transformed by JonesMatrices
# Any matrix can be created with JonesMatrix(A,B,C,D)
# but there are many predefined matrices:
#
# HorizontalPolarizer(): polarizer at theta=0°
# VerticalPolarizer(): polarizer at theta=90° 
# Plus45Polarizer(): polarizer at theta=45°
# Minus45Polarizer(): polarizer at theta=-45°
# HWP(theta=pi/4) : halfwave plate at 45°
# QWP(theta=pi/4) : quarterwave plate at 45° 
# RightCircularPolarizer(): right circular polarizer
# LeftCircularPolarizer(): left circular polarizer
# PhaseRetarder(): arbitrary retarder

print("\n\nTransform the JonesVector with JonesMatrices")
print("============================================")
print("horizontal vector going through quarter waveplate")

vIn = JonesVector.horizontal()
v = QWP(theta=pi/4)*vIn

print("Input components are {0}".format(vIn))
print("Output components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
print("isCircular {0}".format(v.isCircularlyPolarized))
v.show()

print("\n\nApply several JonesMatrices sequentially")
print("============================================")

vIn = JonesVector.horizontal()
v = HWP(theta=pi/2)*QWP(theta=pi/3)*vIn

print("Input components are {0}".format(vIn))
print("Output components are {0}".format(v))
print("Orientation is {0:.2f} rad or {1:.1f}°".format(v.orientation,v.orientation*degPerRad))
print("v.isLinearlyPolarized        = {0}".format(v.isLinearlyPolarized))
print("v.isEllipticallyPolarized    = {0}".format(v.isEllipticallyPolarized))
print("v.isCircularlyPolarized      = {0}".format(v.isCircularlyPolarized))
print("v.isRightCircularlyPolarized = {0}".format(v.isRightCircularlyPolarized))
print("v.isLeftCircularlyPolarized  = {0}".format(v.isLeftCircularlyPolarized))
v.show()
""")


import numpy as np
import math
import sys

class Vector:
    def __init__(self, x: float = 0, y: float = 0, z: float = 0):
        """ Access properties with x,y and z.
        Internally for speed, read access done with self._x """

        if isinstance(x, (int, float)):
            self._x = x
            self._y = y 
            self._z = z
        elif isinstance(x, Vector):
            self._x = x.x
            self._y = x.y 
            self._z = x.z 
        else:
            raise ValueError("No valid input for Vector")

    @property
    def x(self):
        return self._x
    
    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    @x.setter
    def x(self, value):
        self._x = value
    
    @y.setter
    def y(self, value):
        self._y = value

    @z.setter
    def z(self, value):
        self._z = value

    @classmethod
    def fromScaledSum(cls, a, b, scale):
        return Vector(a.x + b.x*scale, a.y + b.y*scale, a.z + b.z*scale)

    def addScaled(self, vector, scale):
        self.x += vector.x*scale
        self.y += vector.y*scale
        self.z += vector.z*scale
        return self

    @property
    def isUnitary(self) -> bool:
        return abs(self.norm()-1) < 1e-7

    @property
    def isNull(self) -> bool:
        return self.norm() < 1e-7

    def __repr__(self):
        return "({0:.4f},{1:.4f},{2:.4f})".format(self.x, self.y, self.z)
    
    def __str__(self):
        return "({0:.4f},{1:.4f},{2:.4f})".format(self.x, self.y, self.z)

    def __mul__(self, scale):
        return Vector(self.x * scale, self.y * scale, self.z * scale)

    def __rmul__(self, scale):
        return Vector(self.x * scale, self.y * scale, self.z * scale)

    def __truediv__(self, scale):
        return Vector(self.x / scale, self.y / scale, self.z / scale)

    def __add__(self, vector):
        return Vector(self.x + vector.x, self.y + vector.y, self.z + vector.z)

    def __neg__(self):
        return Vector(-self.x, -self.y, -self.z)

    def __sub__(self, vector):
        return Vector(self.x - vector.x, self.y - vector.y, self.z - vector.z)

    def __getitem__(self, index):
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        elif index == 2:
            return self.z
        else:
            raise ValueError("Out of range index: must be 0,1 or 2")

    def __eq__(self, vector):
        return self.isEqualTo(vector)

    def isEqualTo(self, vector):
        if self.x != vector.x:
            return False
        if self.y != vector.y:
            return False
        if self.z != vector.z:
            return False
        return True

    def isAlmostEqualTo(self, vector, epsilon=1e-6):
        if abs(self._x - vector.x) > epsilon:
            return False
        if abs(self._y - vector.y) > epsilon:
            return False
        if abs(self._z - vector.z) > epsilon:
            return False
        return True

    def isParallelTo(self, vector, epsilon=1e-7):
        return self.normalizedCrossProduct(vector).abs() < epsilon

    def isPerpendicularTo(self, vector, epsilon=1e-7):
        return abs(self.normalizedDotProduct(vector)) < epsilon

    def anyPerpendicular(self):
        if self._x == 0 and self.y == 0:
            if self._z == 0:
                return None
            else:
                return self.cross(yHat)
        return self.cross(xHat)

    def anyUnitaryPerpendicular(self):
        return self.anyPerpendicular().normalized()

    def isInXYPlane(self, atZ, epsilon=0.001) -> bool:
        if abs(self._z-atZ) < epsilon:
            return True
        return False

    def isInYZPlane(self, atX, epsilon=0.001) -> bool:
        if abs(self._x-atX) < epsilon:
            return True
        return False

    def isInZXPlane(self, atY, epsilon=0.001) -> bool:
        if abs(self._y-atY) < epsilon:
            return True
        return False

    def isInPlane(self, origin: 'Vector', normal: 'Vector', epsilon=0.001) -> bool:
        local = self-origin
        if abs(local.normalizedDotProduct(normal)) < epsilon:
            return True
        return False

    def norm(self):
        ux = self._x
        uy = self._y
        uz = self._z
        return ux*ux+uy*uy+uz*uz

    def abs(self):
        ux = self._x
        uy = self._y
        uz = self._z
        return (ux*ux+uy*uy+uz*uz)**(0.5)

    def normalize(self):
        ux = self._x
        uy = self._y
        uz = self._z

        norm = ux*ux+uy*uy+uz*uz
        if norm != 0:
            invLength = norm**(-0.5)
            self.x *= invLength
            self.y *= invLength
            self.z *= invLength
        return self

    def normalized(self):
        v = Vector(self.x, self.y, self.z)
        return v.normalize()

    def cross(self, vector):
        """ Accessing properties is costly when done very often.
        cross product is a common operation """
        ux = self._x
        uy = self._y
        uz = self._z
        vx = vector.x
        vy = vector.y
        vz = vector.z
        return Vector(uy*vz - uz*vy, uz*vx - ux*vz, ux*vy - uy*vx)

    def dot(self, vector):
        return self.x*vector.x + self.y*vector.y + self.z*vector.z 

    def normalizedCrossProduct(self, vector) -> 'Vector':
        """ Computes the normalized cross product with another vector.
        The absolute value is the sin of the angle between both. You cannot
        get the sign of this angle without providing a perpendicular vector
        that defines the "positive" direction. If you want the angle, 
        use the method angleWith().

        It is twice as fast to use x**(-0.5) rather than 1/sqrt(x).
        """
        productNorm = self.norm() * vector.norm()
        if productNorm == 0:
            return Vector(0,0,0)

        return Vector(self.cross(vector) * (productNorm**(-0.5)))

    def normalizedDotProduct(self, vector):
        """ Computes the normalized dot product with another vector.
        The value is the cos of the angle between both.

        It is twice as fast to use x**(-0.5) rather than 1/sqrt(x)
        """
        productNorm = self.norm() * vector.norm()
        if productNorm == 0:
            return 0
        return self.dot(vector) * (productNorm**(-0.5))

    def angleWith(self, v, axis):
        """ Provides angle 
        """

        sinPhi = self.normalizedCrossProduct(v)
        sinPhiAbs = sinPhi.abs()
        phi = math.asin(sinPhiAbs)
    
        if self.dot(v) <= 0:
            phi = math.pi-phi

        if sinPhi.dot(axis) <= 0:
            phi = -phi
    
        return phi

    def sinAngleWith(self, v, axis):
        """ Provides sine of angle, often sufficient in calculations
        """

        sinPhi = self.normalizedCrossProduct(v)
    
        if sinPhi.dot(axis) <= 0:
            return -sinPhi.abs()
    
        return sinPhi.abs()

    def planeOfIncidence(self, normal):
        if self.dot(normal) < 0:
            normal = -normal

        planeOfIncidenceNormal = self.cross(normal)
        if planeOfIncidenceNormal.norm() < 1e-7:
            return self.anyUnitaryPerpendicular()
        else:
            return planeOfIncidenceNormal.normalized()

    def angleOfIncidence(self, normal):
        if self.dot(normal) < 0:
            normal = -normal

        planeNormal = self.planeOfIncidence(normal)
        return self.angleWith(normal, axis=planeNormal), planeNormal, normal

    def rotateAround(self, u, theta):
        # This is the most expensive (and most common)
        # operation when performing Monte Carlo in tissue 
        # (15% of time spent here). It is difficult to optimize without
        # making it even less readable than it currently is
        # http://en.wikipedia.org/wiki/Rotation_matrix
        # 
        # Several options were tried in the past such as 
        # external not-so-portable C library, unreadable 
        # shortcuts, sine and cosine lookup tables, etc... 
        # and the performance gain was minimal (<20%).
        # For now, this is the best, most readable solution.

        u.normalize()

        cost = math.cos(theta)
        sint = math.sin(theta)
        one_cost = 1 - cost
        
        ux = u._x
        uy = u._y
        uz = u._z
        
        X = self._x
        Y = self._y
        Z = self._z

        self.x = (cost + ux * ux * one_cost) * X \
                 + (ux * uy * one_cost - uz * sint) * Y \
                 + (ux * uz * one_cost + uy * sint) * Z
        self.y = (uy * ux * one_cost + uz * sint) * X \
                 + (cost + uy * uy * one_cost) * Y \
                 + (uy * uz * one_cost - ux * sint) * Z
        self.z = (uz * ux * one_cost - uy * sint) * X \
                 + (uz * uy * one_cost + ux * sint) * Y \
                 + (cost + uz * uz * one_cost) * Z
        return self


class UnitVector(Vector):
    def __init__(self, x: float = 0, y: float = 0, z: float = 0):
        Vector.__init__(self, Vector(x, y, z).normalized())

    def normalizedCrossProduct(self, vector):
        if isinstance(vector, UnitVector):
            return self.cross(vector)
        else:
            return Vector.normalizedCrossProduct(self, vector)

    def normalizedDotProduct(self, vector):
        if isinstance(vector, UnitVector):
            return self.dot(vector)
        else:
            return Vector.normalizedDotProduct(self, vector)


class ConstVector(Vector):
    def __init__(self, x: float = 0, y: float = 0, z: float = 0):
        self._abs = 0
        self._norm = 0
        Vector.__init__(self, x, y, z)
        self._abs = self.abs()
        self._norm = self.norm()

    def normalize(self):
        if self._norm != 1:
            raise RuntimeError("You cannot normalize a constant vector: you can use Vector instead.")
        else:
            raise RuntimeError("You cannot normalize a constant vector: you can use ConstUnitVector instead for unit vectors, there is no need to normalize them.")

    def norm(self):
        return self._norm

    def abs(self):
        return self._abs

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    @x.setter
    def x(self, value):
        raise RuntimeError("You cannot change a constant vector")

    @y.setter
    def y(self, value):
        raise RuntimeError("You cannot change a constant vector")

    @z.setter
    def z(self, value):
        raise RuntimeError("You cannot change a constant vector")

class ConstUnitVector(UnitVector):
    def __init__(self, x: float = 0, y: float = 0, z: float = 0):
        Vector.__init__(self, x, y, z)
        if self.norm() != 1.0:
            raise ValueError("Vector must be created with proper normalized values")

    def normalize(self):
        return self

    def norm(self):
        return 1.0

    def abs(self):
        return 1.0

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    @x.setter
    def x(self, value):
        raise RuntimeError("You cannot change a constant vector")

    @y.setter
    def y(self, value):
        raise RuntimeError("You cannot change a constant vector")

    @z.setter
    def z(self, value):
        raise RuntimeError("You cannot change a constant vector")

oHat = ConstVector(0, 0, 0)    
xHat = ConstUnitVector(1, 0, 0)
yHat = ConstUnitVector(0, 1, 0)
zHat = ConstUnitVector(0, 0, 1)




from numpy import complex, exp, array, sqrt, cos, sin, arctan2, pi, conj, abs, angle
from .utils import *
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as animation

class JonesVector:    
    def __init__(self, Ex: complex = 0.0, Ey: complex = 0.0):
        """ The phase of the field is (k*z-omega*t+phi).
        A positive phase phi is a delayed field, and a negative
        phase is an advanced field.

        See https://en.wikipedia.org/wiki/Jones_calculus

        """
        self.Ex = complex(Ex)
        self.Ey = complex(Ey)
        self.z = 0

    def setValue(self, name, value):
        try:
            setattr(self, name, value)
        except:
            print("Some properties are not mutable")

    def value(self, name):
        return getattr(self, name)

    @property
    def b1(self):
        """ The basis vector for Ex.  It should really be called E1, but
        this is too confusing.  Then b1 should be called bx, but it will not
        always be x̂. For now this is not modifiable. """

        return array([1,0]) # x̂
    
    @property
    def b2(self):
        """ The basis vector for Ey.  It should really be called E2, but
        this is too confusing.  Then b2 should be called by, but it will not
        always be ŷ. For now this is not modifiable. """

        return array([0,1]) # ŷ

    def normalize(self):
        """ Normalize the field amplitudes to obtain an intensity of 1 """

        fieldAmplitude = sqrt(self.intensity)
        if fieldAmplitude != 0:
            self.Ex /= fieldAmplitude
            self.Ey /= fieldAmplitude
        return self

    @property
    def orientation(self):
        """ Orientation of the polarization ellipse.
        Obtained from: https://en.wikipedia.org/wiki/Jones_calculus#Polarization_axis_from_Jones_vector
        """
        Eox = abs(self.Ex)
        Eoy = abs(self.Ey)
        phix = angle(self.Ex)
        phiy = angle(self.Ey)

        x = 2*Eox*Eoy*cos(phix-phiy)
        y = (Eox*Eox-Eoy*Eoy)

        return arctan2(x,y)/2

    @property
    def isLinearlyPolarized(self) -> bool :
        """ The beam is linearly polarized if the phase between both components
        is 0 or pi """
        delta = (angle(self.Ex) - angle(self.Ey)) % pi

        if isAlmostZero(delta):
            return True

        return False
    
    @property
    def isEllipticallyPolarized(self) -> bool :
        return not (self.isLinearlyPolarized or self.isCircularlyPolarized)

    @property
    def isCircularlyPolarized(self) -> bool :
        return self.isRightCircularlyPolarized or self.isLeftCircularlyPolarized

    @property
    def isRightCircularlyPolarized(self) -> bool :
        # See definition of RCP and LCP at https://en.wikipedia.org/wiki/Jones_calculus
        if self.Ey == 0:
            return False

        delta = angle(self.Ex / self.Ey)
        if areRelativelyAlmostEqual(delta, pi / 2):
            if areRelativelyAlmostEqual(abs(self.Ex), abs(self.Ey)):
                return True

        return False

    @property
    def isLeftCircularlyPolarized(self) -> bool :
        if self.Ey == 0:
            return False

        delta = angle(self.Ex / self.Ey)
        if areRelativelyAlmostEqual(delta, -pi / 2):
            if areRelativelyAlmostEqual(abs(self.Ex), abs(self.Ey)):
                return True

        return False

    @property
    def intensity(self) -> float:
        return self.S0

    @property
    def S0(self) -> float:
        return (self.Ex * conj(self.Ex)).real + (self.Ey * conj(self.Ey)).real

    @property
    def S1(self) -> float:
        return (self.Ex * conj(self.Ex)).real - (self.Ey * conj(self.Ey)).real

    @property
    def S2(self) -> float:
        return (self.Ex * conj(self.Ey)).real + (self.Ey * conj(self.Ex)).real

    @property
    def S3(self) -> float:
        raise NotImplementedError("Check math for Stokes vector S3 component")
        
    @property
    def StokesVector(self) -> (float, float, float, float):
        return (self.S0, self.S1, self.S2, self.S3)

    def __str__(self):
        description = ""
        if isEssentiallyReal(self.Ex):
            description += "Ex = {0:.2f}, ".format(abs(self.Ex))
        elif areRelativelyAlmostEqual(abs(self.Ex), 1.0):
            description += "Ex = exp({0}j), ".format(angleInPiFraction(self.Ex))
        else:
            description += "Ex = {0:.2f} ⨉ exp({1}j), ".format(abs(self.Ex), angleInPiFraction(self.Ex))
        
        if isEssentiallyReal(self.Ey):
            description += "Ey = {0:.2f}".format(abs(self.Ey))
        elif areRelativelyAlmostEqual(abs(self.Ey), 1.0):
            description += "Ey = exp({0}j)".format(angleInPiFraction(self.Ey))
        else:
            description += "Ey = {0:.2f} ⨉ exp({1}j)".format(abs(self.Ey), angleInPiFraction(self.Ey))

        return description

    def physicalField(self, phase=0):
        """ The actual physical field that can be measured in the lab (not the complex one)"""
        complexField = self.b1 * self.Ex * exp(1j*phase) + self.b2 * self.Ey * exp(1j*phase)
        return (complexField[0].real, complexField[1].real)

    def realField(self, phase=0):
        """ Synonym of physical field """
        return self.physicalField(phase=phase)

    def fullCycle(self):
        """ A list of points representing the electric field during one complete
        cycle. This is used to display the electric field in self.show() """

        cycle=[]
        for i in range(100):
            phi = 2*pi*(i/99)
            cycle.append( self.physicalField(phase=phi) ) 
        return cycle

    def show(self, filename=None):
        """Animate the electric field on a plot. The arrow represents the electric
        field at any given time. The dashed line is the complete revolution 
        during one cycle. If a filename is provided, it will be saved."""

        cycle = self.fullCycle()
        x,y = zip(*cycle)

        fig, ax = plt.subplots()
        ax.set_ylim(-1,1)
        ax.set_xlim(-1,1)
        ax.set_aspect(1)

        ax.plot(x,y,'k--')
        ax.plot(0,0,'go',markersize=12)

        def animate(point):
            if len(ax.patches) > 0:
                ax.patches.pop(0)
            patch = plt.Arrow(0,0, point[0], point[1], width=0.1, color='k' )
            ax.add_patch(patch)

            return patch,

        ani = animation.FuncAnimation( fig, animate, frames=cycle, interval=30)
        
        if filename is not None:
            ani.save(filename)

        plt.show()

    @classmethod
    def at(cls, theta, inDegrees=False):
        if inDegrees:
            return JonesVector(cos(theta*radPerDeg), sin(theta*radPerDeg))
        return JonesVector(cos(theta), sin(theta))

    @classmethod
    def horizontal(cls):
        return JonesVector.at(theta=0)

    @classmethod
    def vertical(cls):
        return JonesVector.at(theta=pi/2)

    @classmethod
    def plus45(cls):
        return JonesVector.at(theta=pi/4)

    @classmethod
    def minus45(cls):
        return JonesVector.at(theta=-pi/4)

    @classmethod
    def rightCircular(cls):
        return JonesVector(1, exp(-1j*pi/2)).normalize()

    @classmethod
    def leftCircular(cls):
        return JonesVector(1, exp(1j*pi/2)).normalize()
      


from numpy import complex, cos, sin, exp, array, pi, angle, matmul
from numpy.linalg import eig, det
from .utils import *
from .jonesvector import JonesVector
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.animation as animation

class JonesMatrix:
    """ A Jones matrix represents any element that can transform polarization. 
    It is a complex 2 x 2 matrix that can transform a Jones vector
    (representing Ex and Ey).
    """
    def __init__(self, A: complex = 1, B: complex = 0, C: complex = 0, D: complex = 1, m=None, physicalLength: float = 0):
        """ We may initialize the matrix with all four elements or with a numpy
        array. It is mandatory to provide a length for the element, and it is
        assumed that the basis for the matrix is x̂ and ŷ.

        The basis is critical if we want to make sure we don't get confused
        It is very often implicitly x and y, but we will track it explicitly
        to avoid problems
        """
        if m is not None:
            self.mOriginal = array(m)
        else:
            self.mOriginal = array([[A,B],[C,D]])

        self.orientation = 0
        self.L = physicalLength

    @property
    def m(self):
        theta = self.orientation
        rotMatrix = array([[cos(theta),sin(theta)],[-sin(theta), cos(theta)]],dtype='complex')
        invRotMatrix = array([[cos(theta),-sin(theta)],[sin(theta), cos(theta)]],dtype='complex')

        return matmul(invRotMatrix, matmul(self.mOriginal, rotMatrix))

    @property
    def b1(self):
        """ The basis vector for x. For now this is not modifiable. """
        return array([1,0]) # x̂
    
    @property
    def b2(self):
        """ The basis vector for y. For now this is not modifiable. """
        return array([0,1]) # ŷ

    def setValue(self, name, value):
        try:
            setattr(self, name, value)
        except:
            print("Some properties are not mutable")

    def value(self, name):
        return getattr(self, name)

    @property
    def A(self):
        return self.m[0,0]
    @property
    def B(self):
        return self.m[0,1]
    @property
    def C(self):
        return self.m[1,0]
    @property
    def D(self):
        return self.m[1,1]

    @property
    def determinant(self):
        return det(self.m)
    
    @property
    def isBirefringent(self) -> bool:
        """ Returns True if it is birefringent.  See birefringence."""

        phi, e1, e2 = self.birefringence
        if isNotZero(phi, epsilon=1e-7):
            return True
        return False

    @property
    def birefringence(self) :
        """ Returns the birefringence parameters of this element. To determine
        these parameters, one must have a diagonalized version of this matrix
        and then determine if the diagonal element have a phase difference.

        To do that, we simply use a theorem of linear algebra: if a matrix is
        diagonalizable, then its eigenvectors are the appropriate basis and the
        associated eigenvalues are on the diagonal.

        TODO: Need to prove that the basis vectors will always be real.
        TODO: check e1 and e2 and get them to match x and y after rotation

        """
        w, v = eig(self.m)
        phi = angle(w[0]) - angle(w[1])
        e1 = [1,0]
        if isEssentiallyReal(v[0][0]):
            e1[0] = v[0][0].real
        if isEssentiallyReal(v[0][1]):
            e1[1] = v[0][1].real

        if isAlmostZero(e1[0]):
            e1[0] = 0
        if isAlmostZero(e1[1]):
            e1[1] = 0

        e2 = [0,1]
        if isEssentiallyReal(v[1][0]):
            e2[0] = v[1][0].real
        if isEssentiallyReal(v[1][1]):
            e2[1] = v[1][1].real
        if isAlmostZero(e2[0]):
            e2[0] = 0
        if isAlmostZero(e2[1]):
            e2[1] = 0

        return phi, e1, e2

    @property
    def isOpticallyActive(self) -> bool:
        return False

    @property
    def diattenuation(self) -> complex:
        return complex(0, 0)

    @property
    def retardance(self) -> float:
        phi, e1, e2 = self.birefringence
        return phi
    
    def __mul__(self, rightSide):
        """Operator overloading allowing easy-to-read matrix multiplication
        with other `JonesMatrix`, with a `JonesVector`.

        For instance, with M1 = JonesMatrix() and M2 = JonesMatrix(), one can write
        M3 = M1*M2. With r = JonesVector(), one can apply the M1 transform to a vector
        with rOut = M1*r

        """
        if isinstance(rightSide, JonesMatrix):
            return self.mul_matrix(rightSide)
        elif isinstance(rightSide, JonesVector):
            return self.mul_vector(rightSide)
        else:
            raise TypeError(
                "Unrecognized right side element in multiply: '{0}'\
                 cannot be multiplied by a JonesMatrix".format(rightSide))

    def mul_matrix(self, rightSideMatrix: 'JonesMatrix'):
        r""" This function is used to combine two elements into a single matrix.


        Parameters
        ----------
        rightSideMatrix : object from Matrix class
            including the 2 x 2 matrix and other properties of an element.

        Returns
        -------
        A matrix with:

        a : float
            Value of the index (1,1) in the 2 x 2 matrix of the combination of the two elements.
        b : float
            Value of the index (2,1) in the 2 x 2 matrix of the combination of the two elements.
        c : float
            Value of the index (1,2) in the 2 x 2 matrix of the combination of the two elements.
        d : float
            Value of the index (2,2) in the 2 x 2 matrix of the combination of the two elements.

        """

        product = JonesMatrix(m=matmul(self.m, rightSideMatrix.m), physicalLength=self.L + rightSideMatrix.L)
        return product

    def mul_vector(self, rightSideVector):
        r"""This function does the multiplication of a vector by a matrix.
        The output shows the propagated ray through the system.
        New position of ray is updated by the physical length of the matrix.

        Parameters
        ----------
        rightSideVector : object from JonesVector class

        Returns
        -------
        outputVector : an object from JonesVector class
            New position of the input vector after passing through the element.

        """

        outputVector = JonesVector()
        outputVector.Ex = self.A * rightSideVector.Ex + self.B * rightSideVector.Ey
        outputVector.Ey = self.C * rightSideVector.Ex + self.D * rightSideVector.Ey
        outputVector.z = self.L + rightSideVector.z

        return outputVector

    def rotatedBy(self, theta):
        """ We return a rotated copy of the optical element of the matrix by theta. 
        For instance, a theta rotation of a horizontal polarizer will be a polarizer
        aligned at theta. """

        return Rotation(theta = -theta)*self*Rotation(theta = theta)

    def rotateBasisBy(self, theta):
        """ We rotate the basis of the matrix by theta. This is different 
        from `rotateEffectBy` because it does not change the effect, only the
        way it is represented. For instance, a 45 rotation of a horizontal
        polarizer will be a polarizer *still* aligned horizontally, but
        a basis x,y that has been rotated to +45 and +135."""
        raise NotImplemented()

    def show(self, input, xObj, xProperty, xRange, yObj, yProperty):
        if xObj is None:
            xObj = self

        x = []
        y = []

        for value in xRange:
            if xObj is None:
                self.setValue(xProperty, value)
            else:
                xObj.setValue(xProperty, value)

            vOut = self*input

            x.append( xObj.value(xProperty))
            if yObj is None:
                y.append( vOut.value(yProperty))
            else:
                y.append( yObj.value(yProperty))
        
        plt.title("{0} versus {1}".format(yProperty, yProperty))
        plt.xlabel("{0}".format(xProperty))
        plt.ylabel("{0}".format(yProperty))
        plt.plot(x,y,'ko')
        plt.show()


    def showOrientationPlot(self, input:JonesVector):
        x = []
        y = []
        for theta in range(0,190,10):
            theMatrix = JonesMatrix(m=self.m)
            theMatrix.orientation = theta*radPerDeg

            vOut = theMatrix*input
            x.append(theta)
            y.append(vOut.intensity)

        plt.title("Intensity versus orientation of element")
        plt.xlabel(r"Rotation $\theta$ of element [°]")
        plt.ylabel(r"Intensity [arb. unit]")
        plt.plot(x,y,'ko')
        plt.show()

    def showPolarizationPlot(self):
        x = []
        y = []
        for theta in range(0,190,10):
            vIn = JonesVector(Ex=cos(theta*radPerDeg), Ey=sin(theta*radPerDeg))
            vOut = self*vIn
            x.append(theta)
            y.append(vOut.intensity)

        plt.title("Intensity versus input polarization orientation (0° is horizontal)")
        plt.xlabel(r"Input polarization orientation [°]")
        plt.xlim(0,180)
        plt.ylabel(r"Intensity [arb. unit]")
        plt.plot(x,y,'ko')
        plt.show()


class HorizontalPolarizer(JonesMatrix):
    def __init__(self):
        JonesMatrix.__init__(self, A=1,B=0,C=0,D=0,physicalLength=0)        

class LinearPolarizer(JonesMatrix):
    def __init__(self, theta):
        c = cos(theta)
        s = sin(theta)
        JonesMatrix.__init__(self, A=c*c,B=c*s,C=c*s,D=s*s,physicalLength=0)

class VerticalPolarizer(JonesMatrix):
    def __init__(self):
        JonesMatrix.__init__(self, A=0,B=0,C=0,D=1,physicalLength=0)        

class Plus45Polarizer(JonesMatrix):
    def __init__(self):
        JonesMatrix.__init__(self, A=0.5,B=0.5,C=0.5,D=0.5,physicalLength=0)        

class Minus45Polarizer(JonesMatrix):
    def __init__(self):
        JonesMatrix.__init__(self, A=0.5,B=-0.5,C=-0.5,D=0.5,physicalLength=0)        

class RightCircularPolarizer(JonesMatrix):
    def __init__(self):
        JonesMatrix.__init__(self, A=0.5,B=0.5j,C=-0.5j,D=0.5,physicalLength=0)        

class LeftCircularPolarizer(JonesMatrix):
    def __init__(self):
        JonesMatrix.__init__(self, A=0.5,B=-0.5j,C=0.5j,D=0.5,physicalLength=0)        

class Rotation(JonesMatrix):
    def __init__(self, theta):
        JonesMatrix.__init__(self, A=cos(theta), B=sin(theta), C=-sin(theta), D=cos(theta), physicalLength=0)

class PhaseRetarder(JonesMatrix):
    def __init__(self, delta=None, phiX=None, phiY=None, theta=0, physicalLength=0):
        if delta is not None:
            JonesMatrix.__init__(self, A=exp(1j * delta), B=0, C=0, D=1, physicalLength=0)
        else:
            JonesMatrix.__init__(self, A=exp(1j * phiX), B=0, C=0, D=exp(1j * phiY), physicalLength=0)

    
class QWP(PhaseRetarder):
    def __init__(self, theta):
        # theta is fast axis with respect to x-axis
        PhaseRetarder.__init__(self, delta=-pi / 2) # Ex is advanced by pi/2, x is fast
        self.orientation = theta

class HWP(PhaseRetarder):
    def __init__(self, theta):
        PhaseRetarder.__init__(self, delta=-pi) # Ex is advanced by pi, x is fast
        self.orientation = theta

class PockelsCell(JonesMatrix):
    def __init__(self, halfwaveVoltage, length):
        self.halfwaveVoltage = halfwaveVoltage
        self.voltage = 0
        JonesMatrix.__init__(self, A=1, B=0, C=0, D=1, physicalLength=length)

    @property
    def m(self):
        cell = PhaseRetarder(delta=self.voltage/self.halfwaveVoltage*pi)
        cell.orientation = self.orientation
        return cell.m

    def showVoltagePlot(self):
        fig, axs = plt.subplots(1, sharex=True)
        fig.suptitle("Pockels cell at {0:.1f}° with horizontal incident polarization".format(self.orientation*degPerRad))

        voltages = list(range(0,self.halfwaveVoltage+10,10))
        yParallel = []
        yCrossed = []
        for voltage in voltages:
            self.voltage = voltage

            vIn = JonesVector.horizontal()
            vOut = HorizontalPolarizer()*self*vIn
            yParallel.append(vOut.intensity)

            vOut = VerticalPolarizer()*self*vIn
            yCrossed.append(vOut.intensity)


        # axs[0].xlim(0, self.halfwaveVoltage)
        # fig.ylabel(r"Intensity [arb. unit]")
        axs.plot(voltages,yParallel,'k|',label="Between parallel polarizers")
        axs.plot(voltages,yCrossed,'k+',label="Between crossed polarizers")
        axs.legend()
        plt.show()


# class Retarder(JonesMatrix):  # fixme: don't know how to call a JonesMatrixFromRetardanceAndDiattenuation
#     def __init__(self, retardance, diattenuation=None):
#         if diattenuation is None:
#             diattenuation = np.zeros(retardance.shape)
#         dim = retardance.shape
#         f = (diattenuation - 1j * retardance) / 2
#         c = np.sqrt(np.sum(f ** 2, axis=0)).reshape(1, -1)
#         sinch = sinh(c) / c
#         sinch[c == 0] = 1
#         jonesMat = array([[1], [0], [0], [1]]) * (cosh(c)) + sinch * (
#                 array([[1], [0], [0], [-1]]) * f[0, :].reshape(1, -1) +
#                 array([[0], [1], [1], [0]]) * f[1, :].reshape(1, -1) +
#                 array([[0], [1j], [-1j], [0]]) * f[2, :].reshape(1, -1))
#         if np.size(retardance) == 3:
#             jonesMat = jonesMat.reshape((2, 2))
#         else:
#             jonesMat = np.squeeze(jonesMat.reshape(2, 2, dim[1], -1))
#         # return jonesMat



r"""A module to manage polarization with Jones and Stokes formalism
"""

import math

from .jonesvector import *
from .jonesmatrix import *
from .utils import *

__version__ = "1.0.4"
__author__ = "Daniel Cote <dccote@cervo.ulaval.ca>"




import math
import warnings
import inspect
import sys
from numpy import pi, angle

""" Two constants: deg and rad to quickly convert to degrees
or radians with angle*degPerRad or angle*radPerDeg """

degPerRad = 180.0 / math.pi
radPerDeg = math.pi / 180.0


def isAlmostZero(value, epsilon=1e-3):
    return abs(value) < epsilon


def isNotZero(value, epsilon=1e-3):
    return abs(value) > epsilon


def areAbsolutelyAlmostEqual(left, right, epsilon=1e-3):
    return abs(left - right) < epsilon


def areRelativelyAlmostEqual(left, right, epsilon=1e-3):
    absDiff = abs(left - right)
    relTol1 = absDiff / abs(left)
    relTol2 = absDiff / abs(right)
    return relTol1 < epsilon or relTol2 < epsilon


def areAbsolutelyNotEqual(left, right, epsilon=1e-3):
    return abs(left - right) > epsilon


def areRelativelyNotEqual(left, right, epsilon=1e-3):
    return not areRelativelyAlmostEqual(left, right, epsilon)

def isEssentiallyReal(value, epsilon=1e-8) :
    if isAlmostZero(value.imag, epsilon=epsilon):
        return True

    return False

def isEssentiallyImaginary(value, epsilon=1e-8):
    if isAlmostZero(value.real, epsilon=epsilon):
        return True

    return False

def angleInPiFraction(value):
    theta = angle(value)
    denominators = [1,2,3,4,6,8]

    if isAlmostZero(theta):
        return "0"

    for den in denominators:
        for num in range(-den*2, den*2+1,1):
            if areAbsolutelyAlmostEqual(theta, num*pi/den):
                if num == 1:
                    return "π/{0}".format(den)
                elif num == -1:
                    return "-π/{0}".format(den)
                elif float(num)/float(den) == num // den:
                    return "π/{0}".format(num/den)
                else:
                    return "{0}π/{1}".format(num, den)

    return "{0:.2f}".format(theta)


def deprecated(reason: str):
    def deprecatedFunc(func):
        def wrapper(*args, **kwargs):
            warnings.warn(reason, DeprecationWarning)
            return func(*args, **kwargs)

        return wrapper

    return deprecatedFunc


def printClassHierarchy(aClass):
    def printAllChilds(aClass):
        for child in aClass.__subclasses__():
            print("\"{0}\" -> \"{1}\"".format(aClass.__name__, child.__name__))
            printAllChilds(child)
    print("# Paste this in the text field of http://www.graphviz.org")
    print("digraph G {")
    print("  rankdir=\"LR\";")
    printAllChilds(aClass)
    print("}")

def printModuleClasses(moduleName):
    for name, obj in inspect.getmembers(sys.modules[moduleName]):
        if inspect.isclass(obj) and obj.__module__.startswith(moduleName):
            print(obj)



import numpy as np
from simulations.utilities import *


class StokesVector:
    def __init__(self, v:JonesVector):
        self.vector = v
        self.z = 0

    @property
    def S0(self):
        return self.vector.S0
    
    @property
    def S1(self):
        return self.vector.S1

    @property
    def S2(self):
        return self.vector.S2

    @property
    def S3(self):
        return self.vector.S3



import setuptools

""" 
To distribute:
=============
rm dist/*; python setup.py sdist bdist_wheel; python -m twine upload dist/* 

"""


setuptools.setup(
    name="polarization",
    version="1.0.4",
    url="https://github.com/DCC-Lab/Polarization",
    author="Daniel Cote",
    author_email="dccote@cervo.ulaval.ca",
    description="Simple Jones vector and Jones matrices manipulation",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    license='MIT',
    keywords='jones vector matrix polarization optics light',
    packages=setuptools.find_packages(),
    install_requires=['matplotlib>=3', 'numpy'],
    python_requires='>=3.6',
    package_data = {
        # If any package contains *.txt or *.rst files, include them:
        '': ['*.png'],
        "doc": ['*.html']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Education',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

        'Operating System :: OS Independent'
    ],
)
